@@ -1,8 +1,11 @@
 import copy
+from abc import ABC, abstractmethod
+from numbers import Number
 
 import numpy as np
 from scipy.integrate import trapz
 from scipy.interpolate import interp2d
+from scipy.special import gamma
 
 
 def complex_to_polar(complex_vals, phase_degrees=False):
@@ -14,8 +17,8 @@ def complex_to_polar(complex_vals, phase_degrees=False):
     complex_vals : array-like
         Complex number values.
     phase_degrees : bool
-        Whether the phase angles should be returned in 'degrees'. If ``False``,
-        'radians' is assumed.
+        Whether the phase angles should be returned in 'degrees' (``True``) or
+        'radians' (``False``).
 
     Returns
     -------
@@ -59,6 +62,100 @@ def polar_to_complex(amp, phase, phase_degrees=False):
     return amp * (np.cos(phase) + 1j * np.sin(phase))
 
 
+def _check_is_similar(*grids, exact_type=True):
+    """
+    Check if grid objects are similar.
+    """
+    grids = list(grids)
+    grid_ref = grids.pop()
+
+    if exact_type:
+        type_ = type(grid_ref)
+
+        def check_type(grid, grid_type):
+            return type(grid) == grid_type
+
+    else:
+        type_ = Grid
+        check_type = isinstance
+
+    for grid_i in grids:
+        if not check_type(grid_i, type_):
+            raise TypeError("Object types are not similar.")
+        elif grid_ref._vals.shape != grid_i._vals.shape:
+            raise ValueError("Grid objects have different shape.")
+        elif np.any(grid_ref._freq != grid_i._freq) or np.any(
+            grid_ref._dirs != grid_i._dirs
+        ):
+            raise ValueError(
+                "Grid objects have different frequency/direction coordinates."
+            )
+        elif grid_ref.wave_convention != grid_i.wave_convention:
+            raise ValueError("Grid objects have different wave conventions.")
+
+
+def multiply(grid1, grid2, output_type="grid"):
+    """
+    Multiply values (element-wise).
+
+    Parameters
+    ----------
+    grid1 : obj
+        Grid object.
+    grid2 : obj
+        Grid object.
+    type_ : str {"spectrum", "rao", "grid"}
+        Output grid type.
+    """
+
+    TYPE_MAP = {
+        "grid": Grid,
+        "rao": RAO,
+        "directional_spectrum": DirectionalSpectrum,
+        "wave_spectrum": WaveSpectrum,
+    }
+
+    if output_type not in TYPE_MAP:
+        raise ValueError("The given `output_type` is not valid.")
+
+    _check_is_similar(grid1, grid2, exact_type=False)
+
+    type_ = TYPE_MAP.get(output_type)
+
+    freq = grid1._freq
+    dirs = grid1._dirs
+    vals = np.multiply(grid1._vals, grid2._vals)
+    convention = grid1.wave_convention
+
+    new = Grid(
+        freq,
+        dirs,
+        vals,
+        freq_hz=False,
+        degrees=False,
+        **convention,
+    )
+
+    return type_.from_grid(new)
+
+
+def _cast_to_grid(grid):
+    """
+    Cast Grid-like object to ``Grid`` type.
+
+    Note that this type conversion may lead to loss of information/functionality
+    for derived classes.
+    """
+    new = Grid(
+        *grid.grid(freq_hz=grid._freq_hz, degrees=grid._degrees),
+        freq_hz=grid._freq_hz,
+        degrees=grid._degrees,
+        **grid.wave_convention,
+    )
+
+    return new
+
+
 class Grid:
     """
     Two-dimentional frequency/(wave)direction grid.
@@ -117,35 +214,27 @@ class Grid:
                 "and ``M=len(dirs)``."
             )
 
-    def __call__(self, freq_hz=None, degrees=None):
+    @classmethod
+    def from_grid(cls, grid):
         """
-        Return a copy of the grid object's frequency/direction coordinates and corresponding
-        values.
+        Construct from a Grid object.
 
         Parameters
         ----------
-        freq_hz : bool
-            If frequencies should be returned in 'Hz'. If ``False``, 'rad/s' is used.
-            Defaults to original units used during initialization.
-        degrees : bool
-            If directions should be returned in 'degrees'. If ``False``, 'radians'
-            is used. Defaults to original units used during initialization.
+        grid : obj
+            Grid object.
 
         Returns
         -------
-        freq : array
-            1-D array of grid frequency coordinates.
-        dirs : array
-            1-D array of grid direction coordinates.
-        vals : array (N, M)
-            Grid values as 2-D array of shape (N, M), such that ``N=len(freq)``
-            and ``M=len(dirs)``.
+        cls :
+            Initialized object.
         """
-        freq = self.freq(freq_hz=freq_hz)
-        dirs = self.dirs(degrees=degrees)
-        vals = self._vals.copy()
-
-        return freq, dirs, vals
+        return cls(
+            *grid.grid(freq_hz=grid._freq_hz, degrees=grid._degrees),
+            freq_hz=grid._freq_hz,
+            degrees=grid._degrees,
+            **grid.wave_convention,
+        )
 
     def _check_freq(self, freq):
         """
@@ -206,6 +295,36 @@ class Grid:
 
         return dirs
 
+    def grid(self, freq_hz=None, degrees=None):
+        """
+        Return a copy of the object's frequency/direction coordinates and corresponding
+        grid values.
+
+        Parameters
+        ----------
+        freq_hz : bool
+            If frequencies should be returned in 'Hz'. If ``False``, 'rad/s' is used.
+            Defaults to original units used during initialization.
+        degrees : bool
+            If directions should be returned in 'degrees'. If ``False``, 'radians'
+            is used. Defaults to original units used during initialization.
+
+        Returns
+        -------
+        freq : array
+            1-D array of grid frequency coordinates.
+        dirs : array
+            1-D array of grid direction coordinates.
+        vals : array (N, M)
+            Grid values as 2-D array of shape (N, M), such that ``N=len(freq)``
+            and ``M=len(dirs)``.
+        """
+        freq = self.freq(freq_hz=freq_hz)
+        dirs = self.dirs(degrees=degrees)
+        vals = self._vals.copy()
+
+        return freq, dirs, vals
+
     @property
     def wave_convention(self):
         """
@@ -376,8 +495,8 @@ class Grid:
         self,
         freq,
         dirs,
-        freq_hz=True,
-        degrees=True,
+        freq_hz=False,
+        degrees=False,
         complex_convert="rectangular",
         fill_value=0.0,
     ):
@@ -416,8 +535,8 @@ class Grid:
         array :
             Interpolated grid values.
         """
-        freq = np.asarray_chkfinite(freq).copy()
-        dirs = np.asarray_chkfinite(dirs).copy()
+        freq = np.asarray_chkfinite(freq).reshape(-1)
+        dirs = np.asarray_chkfinite(dirs).reshape(-1)
 
         if freq_hz:
             freq = 2.0 * np.pi * freq
@@ -438,8 +557,8 @@ class Grid:
         self,
         freq,
         dirs,
-        freq_hz=True,
-        degrees=True,
+        freq_hz=False,
+        degrees=False,
         complex_convert="rectangular",
         fill_value=0.0,
     ):
@@ -503,77 +622,139 @@ class Grid:
 
     def __mul__(self, other):
         """
-        Multiply values with another Grid object.
+        Multiply values (element-wise).
 
-        Both grids must have the same frequency/direction coordinates.
+        Both grids must have the same frequency/direction coordinates, and the same
+        'wave convention'.
 
         Parameters
         ----------
-        other : obj
-            Grid object to be multiplied with.
+        other : obj or numeric
+            Grid object or number to be multiplied with.
 
         Returns
         -------
         obj :
-            A copy of the object where the values are multiplied with another Grid.
-        """
-        if not isinstance(other, Grid):
-            raise ValueError()
-        elif self._vals.shape != other._vals.shape:
-            raise ValueError()
-        elif np.any(self._freq != other._freq) or np.any(self._dirs != other._dirs):
-            raise ValueError()
-        elif self.wave_convention != other.wave_convention:
-            raise ValueError()
-
-        new = Grid(
-            self._freq,
-            self._dirs,
-            self._vals * other._vals,
-            freq_hz=False,
-            degrees=False,
-            **self.wave_convention,
-        )
-
-        if isinstance(self, DirectionalSpectrum) or isinstance(
-            other, DirectionalSpectrum
-        ):
-            return DirectionalSpectrum.from_grid(new)
-
-        return new
-
-    def __abs__(self):
-        """
-        Return new object where values are converted to absolute values.
+            A copy of the object where the values are multiplied with values of
+            another grid.
         """
         new = self.copy()
-        new._vals = np.abs(new._vals)
+
+        if isinstance(other, Number):
+            new._vals = new._vals * other
+        else:
+            _check_is_similar(self, other, exact_type=True)
+            new._vals = new._vals * other._vals
+
         return new
 
+    def __rmul__(self, other):
+        return self.__mul__(other)
+
+    def __add__(self, other):
+        """
+        Add values (element-wise).
+
+        Both grids must have the same frequency/direction coordinates, and the same
+        'wave convention'.
+
+        Parameters
+        ----------
+        other : obj or numeric
+            Grid object or number to be added.
+
+        Returns
+        -------
+        obj :
+            A copy of the object where the values are added with another grid's values.
+        """
+        new = self.copy()
+
+        if isinstance(other, Number):
+            new._vals = new._vals + other
+        else:
+            _check_is_similar(self, other, exact_type=True)
+            new._vals = new._vals + other._vals
+
+        return new
+
+    def __radd__(self, other):
+        return self.__add__(other)
+
+    def __sub__(self, other):
+        """
+        Subtract values (element-wise).
+
+        Both grids must have the same frequency/direction coordinates, and the same
+        'wave convention'.
+
+        Parameters
+        ----------
+        other : obj or numeric
+            Grid object or number to be subtracted.
+
+        Returns
+        -------
+        obj :
+            A copy of the object where the values are subtracted with another grid's values.
+        """
+        new = self.copy()
+
+        if isinstance(other, Number):
+            new._vals = new._vals - other
+        else:
+            _check_is_similar(self, other, exact_type=True)
+            new._vals = new._vals - other._vals
+
+        return new
+
+    def __rsub__(self, other):
+        return self.__sub__(other)
+
     def __repr__(self):
         return "Grid"
 
+    def conjugate(self):
+        """
+        Return a copy of the object with complex conjugate values.
+        """
+        new = self.copy()
+        new._vals = new._vals.conjugate()
+        return new
+
     @property
     def real(self):
         """
-        Return a copy of the object where all values are converted to their real
-        part.
+        Return a new Grid object where all values are converted to their real part.
         """
-        new = self.copy()
+        new = _cast_to_grid(self)
         new._vals = new._vals.real
         return new
 
     @property
     def imag(self):
         """
-        Return a copy of the object where all values are converted to their imaginary
-        part.
+        Return a new Grid object where all values are converted to their imaginary part.
         """
-        new = self.copy()
+        new = _cast_to_grid(self)
         new._vals = new._vals.imag
         return new
 
 
+class DisableComplexMixin:
+    @property
+    def imag(self):
+        raise AttributeError(f"'{self}' object has no attribute 'imag'.")
+
+    @property
+    def real(self):
+        raise AttributeError(f"'{self}' object has no attribute 'real'.")
+
+    @property
+    def conjugate(self):
+        raise AttributeError(f"'{self}' object has no attribute 'conjugate'.")
+
+
 class RAO(Grid):
     """
     Response amplitude operator (RAO).
@@ -689,14 +870,6 @@ class RAO(Grid):
         rao._phase_degrees = phase_degrees
         return rao
 
-    def conjugate(self):
-        """
-        Return a copy of the object with complex conjugate values.
-        """
-        new = self.copy()
-        new._vals = new._vals.conjugate()
-        return new
-
     def differentiate(self, n=1):
         """
         Return the nth derivative of the RAO.
@@ -751,25 +924,15 @@ class RAO(Grid):
         if phase_degrees is None:
             phase_degrees = self._phase_degrees
 
-        freq, dirs, vals = self(freq_hz=freq_hz, degrees=degrees)
+        freq, dirs, vals = self.grid(freq_hz=freq_hz, degrees=degrees)
         vals_amp, vals_phase = complex_to_polar(vals, phase_degrees=phase_degrees)
         return freq, dirs, vals_amp, vals_phase
 
-    def __abs__(self):
-        return Grid(
-            self._freq,
-            self._dirs,
-            np.abs(self._vals),
-            freq_hz=False,
-            degrees=False,
-            **self.wave_convention,
-        )
-
     def __repr__(self):
         return "RAO"
 
 
-class DirectionalSpectrum(Grid):
+class DirectionalSpectrum(DisableComplexMixin, Grid):
     """
     Directional spectrum.
 
@@ -823,10 +986,10 @@ class DirectionalSpectrum(Grid):
         )
 
         if freq_hz:
-            self._vals /= 2.0 * np.pi
+            self._vals = 1.0 / (2.0 * np.pi) * self._vals
 
         if degrees:
-            self._vals /= np.pi / 180.0
+            self._vals = 180.0 / np.pi * self._vals
 
         if np.any(np.iscomplex(self._vals)):
             raise ValueError("Spectrum values can not be complex.")
@@ -834,37 +997,86 @@ class DirectionalSpectrum(Grid):
             raise ValueError("Spectrum values must be positive.")
 
     @classmethod
-    def from_grid(cls, grid):
+    def from_spectrum1d(
+        cls,
+        freq,
+        dirs,
+        spectrum1d,
+        spread_fun,
+        dirp,
+        freq_hz=False,
+        degrees=False,
+        clockwise=False,
+        waves_coming_from=True,
+    ):
         """
-        Construct a ``DirectionalSpectrum`` object from a ``Grid`` object.
+        Construct a 2-D 'directional' spectrum from a 1-D 'non-directional' spectrum,
+        a spreading function and a peak direction.
 
-        It is assumed that the grid's values represent spectrum density values that
-        correspond to frequency/direction units given during initialization. Note
-        that if you scale the frequency/direction bins of the spectrum, you must
-        also scale the corresponding spectrum density values (since the total energy/integral
-        of the spectrum should be preserved).
+        The directional spectrum is constructed according to:
+
+            ``S(f, theta) = S(f) * D(f, theta - theta_p)``
+
+        where ``S(f)`` is the non-directional spectrum, ``D(f, theta - theta_p)``
+        is the directional spreading function, and ``theta_p`` is the peak direction.
+        ``f`` is the frequency coordinate, and ``theta`` is the direction coordinate.
 
         Parameters
         ----------
-        grid : obj
-            Grid object.
-
-        Returns
-        -------
-        cls :
-            Initialized DirectionalSpectrum object.
+        freq : array-like
+            1-D array of grid frequency coordinates. Positive and monotonically increasing.
+        dirs : array-like
+            1-D array of grid direction coordinates. Positive and monotonically increasing.
+            Must cover the directional range [0, 360) degrees (or [0, 2 * numpy.pi) radians).
+        spectrum1d : array-like
+            1-D array of non-directional spectrum density values. These 1-D spectrum
+            values will be scaled according to the spreading function, and distributed
+            to all frequency/direction coordinates. `spectrum1d` must have the same
+            length as `freq`.
+        spread_fun : callable
+            Spreading function. Takes a frequency coordinate (float) and a direction
+            coordinate (float) as input, and returns a corresponding scaling value
+            (float).
+        dirp : float
+            Peak direction. Direction in which the spectrum has its maximum values.
+        freq_hz : bool
+            If frequency is given in 'Hz'. If ``False``, 'rad/s' is assumed.
+        degrees : bool
+            If direction is given in 'degrees'. If ``False``, 'radians' is assumed.
+        clockwise : bool
+            If positive directions are defined to be 'clockwise'. If ``False``,
+            'counterclockwise' is assumed.
+        waves_coming_from : bool
+            If waves are 'coming from' the given directions. If ``False``, 'going towards'
+            convention is assumed.
         """
+        spectrum1d = np.asarray_chkfinite(spectrum1d).reshape(-1, 1)
+        vals = np.tile(spectrum1d, (1, len(dirs)))
+
+        if degrees:
+            period = 360.0
+        else:
+            period = 2.0 * np.pi
+
+        for (idx_f, idx_d), val_i in np.ndenumerate(vals):
+            f_i = freq[idx_f]
+            d_i = (dirs[idx_d] - dirp) % period
+            vals[idx_f, idx_d] = spread_fun(f_i, d_i) * val_i
+
         return cls(
-            *grid(freq_hz=grid._freq_hz, degrees=grid._degrees),
-            freq_hz=grid._freq_hz,
-            degrees=grid._degrees,
-            **grid.wave_convention,
+            freq,
+            dirs,
+            vals,
+            freq_hz=freq_hz,
+            degrees=degrees,
+            clockwise=clockwise,
+            waves_coming_from=waves_coming_from,
         )
 
     def __repr__(self):
         return "DirectionalSpectrum"
 
-    def __call__(self, freq_hz=False, degrees=False):
+    def grid(self, freq_hz=False, degrees=False):
         """
         Return a copy of the spectrum's frequency/direction coordinates and corresponding
         values.
@@ -887,7 +1099,7 @@ class DirectionalSpectrum(Grid):
             Spectrum density values as 2-D array of shape (N, M), such that ``N=len(freq)``
             and ``M=len(dirs)``.
         """
-        freq, dirs, vals = super().__call__(freq_hz=freq_hz, degrees=degrees)
+        freq, dirs, vals = super().grid(freq_hz=freq_hz, degrees=degrees)
 
         if freq_hz:
             vals *= 2.0 * np.pi
@@ -901,8 +1113,8 @@ class DirectionalSpectrum(Grid):
         self,
         freq,
         dirs,
-        freq_hz=True,
-        degrees=True,
+        freq_hz=False,
+        degrees=False,
         fill_value=0.0,
         **kwargs,
     ):
@@ -1014,7 +1226,7 @@ class DirectionalSpectrum(Grid):
         elif degrees is None:
             degrees = self._degrees
 
-        freq, dirs, vals = self(freq_hz=freq_hz, degrees=degrees)
+        freq, dirs, vals = self.grid(freq_hz=freq_hz, degrees=degrees)
 
         if axis == 0:
             y = freq
@@ -1053,6 +1265,9 @@ class DirectionalSpectrum(Grid):
 
 
 class WaveSpectrum(DirectionalSpectrum):
+    def __repr__(self):
+        return "WaveSpectrum"
+
     @property
     def hs(self):
         """
@@ -1074,7 +1289,7 @@ class WaveSpectrum(DirectionalSpectrum):
 
             ``tz = np.sqrt(m0 / m2)``
         """
-        m0 = self.moment(0)
+        m0 = self.moment(0, freq_hz=False)
         m2 = self.moment(2, freq_hz=True)
         return np.sqrt(m0 / m2)
 
@@ -1216,4 +1431,148 @@ def calculate_response(
     rao_squared = rao_squared.reshape(freq, dirs, freq_hz=False, degrees=False)
     wave_body = wave_body.reshape(freq, dirs, freq_hz=False, degrees=False)
 
-    return rao_squared * wave_body
+    return multiply(rao_squared, wave_body, output_type="directional_spectrum")
+
+
+class BaseSpreading(ABC):
+    """
+    Base class for spreading functions.
+
+    Parameters
+    ----------
+    freq_hz : bool
+        If frequencies passed to the spreading function will be given in 'Hz'.
+        If ``False``, 'rad/s' is assumed.
+    degrees : bool
+        If directions passed to the spreading function will be given in 'degrees'.
+        If ``False``, 'radians' is assumed.
+    """
+
+    def __init__(self, freq_hz=False, degrees=False):
+        self._freq_hz = freq_hz
+        self._degrees = degrees
+
+    def __call__(self, frequency, direction):
+        """
+        Get spreading value for given frequency/direction coordinate.
+
+        Parameters
+        ----------
+        frequency : float
+            Frequency coordinate. Units should be according to the `freq_hz` flag
+            given during initialization.
+        direction : float
+            Direction coordinate. Units should be according to the `degrees` flag
+            given during initialization.
+        """
+        if self._freq_hz:
+            frequency = 2.0 * np.pi * frequency
+
+        if self._degrees:
+            direction = (np.pi / 180.0) * direction
+            scale = 1.0 / 360.0
+        else:
+            scale = 1.0 / (2.0 * np.pi)
+
+        return scale * self._spread_fun(frequency, direction % (2.0 * np.pi))
+
+    @abstractmethod
+    def _spread_fun(self, omega, theta):
+        """
+        Get spreading value for given frequency/direction coordinate.
+
+        Parameters
+        ----------
+        omega : float
+            Frequency coordinate in 'rad/s'.
+        direction : float
+            Direction coordinate in 'radians'.
+        """
+        raise NotImplementedError()
+
+
+class CosineHalfSpreading(BaseSpreading):
+    """
+    Cosine-2s type spreading (half directional range).
+
+    Defined as:
+
+        ``D(theta) = scale * C(s) * cos(theta) ** (2 * s)`` , for -pi/2 <= theta <= pi/2
+
+        ``D(theta) = 0`` , otherwise
+
+    where,
+
+        ``C(s) = 2 ** (2 * s + 1) * gamma(s + 1) ** 2 / gamma(2 * s + 1)``
+
+    If `theta` is given in 'radians':
+
+        ``scale = 1.0 / (2.0 * np.pi)``
+
+    If `theta` is given in 'degrees':
+
+        ``scale = 1 / 360.0``
+
+    Note that this spreading is independent of frequency.
+
+    Parameters
+    ----------
+    s : int
+        Spreading coefficient.
+    degrees : bool
+        If directions passed to the spreading function will be given in 'degrees'.
+        If ``False``, 'radians' is assumed.
+    """
+
+    def __init__(self, s=1, degrees=False):
+        self._s = s
+        super().__init__(degrees=degrees)
+
+    def _spread_fun(self, _, theta, /):
+        if (np.pi / 2.0) <= theta <= (3.0 * np.pi / 2.0):
+            return 0
+
+        s = self._s
+        c = 2 ** (2 * s + 1) * gamma(s + 1) ** 2 / gamma(2 * s + 1)
+        return c * (np.cos(theta) ** 2.0) ** s
+
+
+class CosineFullSpreading(BaseSpreading):
+    """
+    Cosine-2s type spreading (full directional range).
+
+    Defined as:
+
+        ``D(theta) = scale * C(s) * cos(theta / 2) ** (2 * s)``
+
+    where,
+
+        ``C(s) = 2 ** (2 * s) * gamma(s + 1) ** 2 / gamma(2 * s + 1)``
+
+    If `theta` is given in 'radians':
+
+        ``scale = 1.0 / (2.0 * np.pi)``
+
+    If `theta` is given in 'degrees':
+
+        ``scale = 1 / 360.0``
+
+    Note that this spreading is independent of frequency.
+
+    Parameters
+    ----------
+    s : int
+        Spreading coefficient.
+    degrees : bool
+        If directions passed to the spreading function will be given in 'degrees'.
+        If ``False``, 'radians' is assumed.
+    """
+
+    def __init__(self, s=1, degrees=False):
+        self._s = s
+        super().__init__(degrees=degrees)
+
+    def _spread_fun(self, _, theta, /):
+        s = self._s
+        c = 2 ** (2 * s) * gamma(s + 1) ** 2 / gamma(2 * s + 1)
+        return c * (np.cos(theta / 2.0) ** 2.0) ** s
